from typing import Optional, Tuple

import pandas as pd
import plotly.graph_objs as go
from dash import Input, Output, callback, dcc, html
from statsmodels.tsa.api import seasonal_decompose
from ts_app import plotting
from ts_app.ts_functions import create_arma_sample, fit_arima_model

model_param_input = html.Div(
    id="model-params",
    className="param-input",
    children=[
        html.H3("Model parameters"),
        # AR order dropdown
        html.Label("AR Order", htmlFor="model-ar"),
        dcc.Dropdown(
            id="model-ar",
            clearable=False,
            placeholder="AR order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(6)],
        ),
        # Differencing order dropdown
        html.Label("Differencing Order", htmlFor="model-diff"),
        dcc.Dropdown(
            id="model-diff",
            clearable=False,
            value=0,
            placeholder="Differencing",
            searchable=False,
            options=[{"label": f"{i}", "value": i} for i in range(6)],
        ),
        # MA order dropdown
        html.Label("MA Order", htmlFor="model-ma"),
        dcc.Dropdown(
            id="model-ma",
            clearable=False,
            placeholder="MA order",
            searchable=False,
            value=1,
            options=[{"label": f"{i}", "value": i} for i in range(6)],
        ),
    ],
)

seasonal_decomposition_plot = dcc.Loading(
    dcc.Graph(id="component-plots"), color="#777"
)

forecast_line_plot = dcc.Loading(
    dcc.Graph(
        id="line-plot", config={"toImageButtonOptions": {"format": "svg"}}
    ),
    color="#777",
)

EXPLANATORY_TEXT = """
### Guide

Graphs reveal the [trend](/glossary#Trend) in the data, and help assess the
**goodness of fit**. In general, a good model should reasonably replicate the
behaviour of the historical data.

Graphs also help discover [seasonal](/glossary#Seasonality) and
[cyclic patterns](/glossary#Cyclic-patterns). These usually manifest as
occasional peaks or troughs.

[Autocorrelation][1] and [Partial-Autocorrelation][2] plots can provide hints
on a potentially suitable model to start with. [This article][3] describes how.
Testing for [stationarity][4], and filtering out seasonal & trend effects is
an essential first step.

[1]: https://en.wikipedia.org/wiki/Autocorrelation
[2]: https://en.wikipedia.org/wiki/Partial_autocorrelation_function
[3]: https://en.wikipedia.org/wiki/Box%E2%80%93Jenkins_method#Autocorrelation_\
and_partial_autocorrelation_plots
[4]: https://cran.r-project.org/web/packages/TSTutorial/vignettes/Stationary\
.pdf

"""

footer_buttons = html.Div(
    [
        html.A("Back to home", href="/", className="button"),
        html.A("Browse glossary", href="/glossary", className="button"),
    ],
    className="footer",
)


def generate_layout(input_source: html.Div) -> html.Div:
    """Get a dashboard layout with the given `input_source`.

    Args:
        input_source (dash.html.Div.Div): A html form to collect input.

    Returns:
        dash.html.Div.Div: Dashboard layout.
    """
    return html.Div(
        id="dashboard-content",
        className="dashboard-content",
        children=[
            # Current path
            dcc.Location(id="current-page"),
            # Sidebar with input forms
            html.Div(
                className="side-bar",
                children=[
                    html.Div(input_source, id="data-source"),
                    dcc.Store(id="sample-data-store"),
                    dcc.Store(id="file-upload-store"),
                    model_param_input,
                ],
            ),
            # Time series forecasting plots and guide
            html.Div(
                className="graphs-and-guide",
                children=[
                    seasonal_decomposition_plot,
                    html.Div(
                        [
                            forecast_line_plot,
                            dcc.Markdown(EXPLANATORY_TEXT, className="guide"),
                            footer_buttons,
                        ],
                        className="line-plot-and-summary",
                    ),
                ],
            ),
        ],
    )


@callback(
    [
        Output("line-plot", "figure"),
        Output("component-plots", "figure"),
    ],
    [
        Input("model-ar", "value"),
        Input("model-diff", "value"),
        Input("model-ma", "value"),
        Input("current-page", "pathname"),
        Input("sample-data-store", "data"),
        Input("file-upload-store", "data"),
    ],
)
def model_and_predict(
    ar_order: int,
    diff_order: int,
    ma_order: int,
    input_source: str,
    sample: Optional[dict],
    upload: Optional[dict],
) -> Tuple[go.Figure, go.Figure]:
    """Fit an ARIMA model each time model parameters or input data are
    modified, then plot the results.

    Args:
        ar_order (int): AR order.
        diff_order (int): Differencing order.
        ma_order (int): MA order.
        input_source (str): The data source.
        sample (Optional[dict]): Stored sample data, if any.
        upload (Optional[dict]): Uploaded data, if any.

    Returns:
        Tuple[Figure, Figure]: A line-plot of the forecast results, and
        subplots with seasonal decomposition estimates.
    """

    if input_source == "/upload" and upload is not None:
        filename = upload["filename"]
        data = pd.read_json(upload["data"], orient="index", typ="series")
    elif input_source == "/sample" and sample is not None:
        filename = sample["filename"]
        data = pd.read_json(sample["data"], orient="index", typ="series")
    else:
        filename = "a random sample"
        data = create_arma_sample()

    predictions, forecast = fit_arima_model(
        data, ar_order, diff_order, ma_order
    )

    line_plot = plotting.plot_forecast(
        actual_data=data,
        predictions=predictions,
        forecast=forecast,
        model_info=f"ARIMA({ar_order}, {diff_order}, {ma_order})",
        file_name=filename,
    )

    components = seasonal_decompose(data)
    component_subplots = plotting.plot_ts_components(
        trend=components.trend,
        seasonal=components.seasonal,
        residuals=components.resid,
        file_name=filename,
    )

    return line_plot, component_subplots



from typing import Optional

import pandas as pd
from dateutil.parser import ParserError as dtParserError


def process_upload(data: pd.DataFrame) -> Optional[str]:
    """Validate the data obtained from an uploaded file.

    Args:
        data (pandas.DataFrame): The data extracted from the file.

    Returns:
        Optional[str]: The error found, if any.
    """
    if (n := len(data)) < 32:
        return (
            f"Please try again... the uploaded file has only {n} values, but"
            " the minimum is set at 32."
        )

    # Ensure the dates(index) can be converted to datetime objects having a
    # consistent frequency.
    # If date frequency can't be inferred, or is not consistent, some
    # statsmodels time series functions won't work.
    try:
        data.index = pd.to_datetime(data.index)

        if pd.infer_freq(data.index) in {None, "N"}:
            return (
                "Please try again... a uniform date frequency "
                "(which is needed in some of the time series functions used) "
                "could not be determined."
            )
    except (dtParserError, TypeError, ValueError):
        return (
            "Please try again... it seems that the values in the 1st column "
            "could not be read as dates."
        )

    # Ensure the data has at least 2 columns, since the first column is read
    # as the index.
    # If multiple columns exist, consider the right-most (last).
    try:
        data = data.select_dtypes(include="number").iloc[:, -1]
    except IndexError:
        return (
            "Please ensure that the data has a date index and at least one "
            "other column with numeric values."
        )



import json
from pkgutil import get_data

import dash
from dash import html

dash.register_page(__name__)

GLOSSARY_DATA = json.loads(get_data("ts_app", "assets/glossary.json"))

title_definition_pairs = sum(  # Concatenate list comprehension items
    [
        [  # Each item is list with a single Div
            html.Div(
                id=item["title"].replace(" ", "-"),
                children=[
                    html.Dt(item["title"]),
                    html.Dd(item["definition"]),
                ],
            )
            for item in sorted(GLOSSARY_DATA, key=lambda x: x["title"])
        ]
    ],
    start=[],
)

layout = html.Div(
    [
        html.Div(
            className="glossary-content",
            children=[
                html.H2("Glossary of Time Series Terms"),
                html.Dl(title_definition_pairs),
            ],
        ),
        html.Div(
            className="footer",
            children=[html.A("Back to home", href="/", className="button")],
        ),
    ]
)



import pandas as pd
from ts_app.file_upload import process_upload

date_index = pd.date_range("2020-01-01", periods=50)

good_sample = pd.DataFrame(range(50), index=date_index)
non_numeric_sample = pd.DataFrame(list("abcde") * 10, index=date_index)
small_sample = pd.DataFrame(range(20), index=date_index[:20])
non_dated_sample = pd.DataFrame(range(50), index=[f"a{i}" for i in range(50)])


def test_valid_upload():
    error = process_upload(good_sample)
    assert error is None


def test_small_size_uploads():
    error = process_upload(small_sample)
    assert "but the minimum is set at 32" in error


def test_non_numeric_uploads():
    error = process_upload(non_numeric_sample)
    assert "at least one other column with numeric values" in error


def test_uploads_without_dates():
    error = process_upload(non_dated_sample)
    assert "could not be read as dates." in error



import sys

from ts_app.cli import process_cli_args


def test_default_args(monkeypatch):
    """Check whether default arguments are set."""
    monkeypatch.setattr(sys, "argv", ["ts_app"])
    args = process_cli_args()

    assert args.host == "localhost"
    assert args.port == 8000
    assert args.no_browser is False


def test_supplied_args(monkeypatch):
    """Check whether input arguments are collected."""
    monkeypatch.setattr(
        sys,
        "argv",
        "ts_app -p 25000 --host https://example.com --no-browser".split(),
    )
    args = process_cli_args()

    assert args.host == "https://example.com"
    assert args.port == 25000
    assert args.no_browser is True



import pandas as pd
from pandas.api.types import is_datetime64_dtype
from ts_app.ts_functions import create_arma_sample, fit_arima_model


def test_default_sample_creation():
    default_sample = create_arma_sample()

    assert isinstance(default_sample, pd.Series)
    assert default_sample.shape == (100,)
    assert is_datetime64_dtype(default_sample.index)


def test_specified_sample_creation():
    default_sample = create_arma_sample(ar_order=2, ma_order=2, size=200)

    assert isinstance(default_sample, pd.Series)
    assert default_sample.shape == (200,)
    assert is_datetime64_dtype(default_sample.index)


def test_ts_prediction():
    data = create_arma_sample()
    predictions, forecast = fit_arima_model(data)

    assert predictions.shape == (31,)
    assert forecast.shape == (15,)
    assert is_datetime64_dtype(predictions.index)



import dash
from dash import Dash, html

app = Dash(
    "ts_app",
    suppress_callback_exceptions=True,
    external_scripts=["https://cdn.plot.ly/plotly-2.22.0.min.js"],
    title="Time Series App",
    meta_tags=[
        {
            "name": "viewport",
            "content": "width=device-width, initial-scale=1.0",
        }
    ],
    use_pages=True,
)

app.layout = html.Div(dash.page_container)



import logging
import sys
import threading
import webbrowser

import waitress

from ts_app.cli import process_cli_args
from ts_app.dash_app import app

__version__ = "0.9.1"

logging.basicConfig(level="INFO")

# Suppress benign queue warnings
logging.getLogger("waitress.queue").setLevel("ERROR")


def custom_hook(type, value, traceback) -> None:
    """Custom exception handler to allow clean stopping of the app server
    using `CTRL + C` i.e. KeyboardInterrupt.

    Args:
        type (Exception): The unhandled exception encountered.
        value (str): The exception's arg(s).
        traceback (traceback): Stack trace object.
    """
    if type is KeyboardInterrupt:
        print("\nServer Stopped.")
        exit()
    else:
        print(type, value)


sys.excepthook = custom_hook


def run_app(
    host: str = "localhost", port: int = 8000, launch_browser: bool = True
) -> None:
    """Start the app server, and launch a web browser to it.

    Args:
        host (str, optional): A host-name/IP address. Defaults to "localhost".
        port (int, optional): TCP port to listen at. Defaults to 8000.
        launch_browser (bool, optional): Whether to launch a web browser to
            view the app. Defaults to True.
    """
    server_ = threading.Thread(
        target=waitress.serve,
        kwargs=dict(app=app.server, host=host, port=port),
    )
    server_.start()

    if launch_browser is True:
        webbrowser.open(f"{host}:{port}")

    server_.join()


def _run_in_cli() -> None:
    """Start the app server using input collected from the command line
    interface.
    """
    args = process_cli_args()

    run_app(
        host=args.host.split("://")[-1],
        port=args.port,
        launch_browser=not args.no_browser,  # True if `no_browser` is not set
    )



import dash
from dash import dcc, html

dash.register_page(__name__, path="/", chapati=777)

HOMEPAGE_TEXT = """
# Time Series App

---

A simple application that illustrates basic [time series analysis][ts-wiki]
concepts. A [glossary](/glossary) of terms is available to help get started.

You can upload a file, or generate sample data, then interactively fit
[ARIMA][arima] models.

In a nutshell, *Time Series Analysis* purposes to learn and emulate the
behaviour of data over time. *Time Series Forecasting* then leverages this
insight to estimate future values.

![An example of time series analysis results, plotted](/assets/ts.svg)

[ts-wiki]: https://en.wikipedia.org/wiki/Time_series
[arima]: https://en.wikipedia.org/wiki/Autoregressive_integrated_moving_\
average
"""

layout = html.Div(
    className="home-page-content",
    children=[
        # Introduction
        dcc.Markdown(HOMEPAGE_TEXT),
        html.Div(
            className="footer",
            children=[
                html.A("Upload a file", href="/upload", className="button"),
                html.A(
                    "Process a sample",
                    href="/sample",
                    className="button",
                ),
            ],
        ),
    ],
)
